/**********************************************************************
Copyright (c) 2011 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2012 Chris Rued - patch for MapReduceCount and collection.count()
   ...
**********************************************************************/
package org.datanucleus.store.mongodb;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.bson.types.ObjectId;
import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.FetchPlan;
import org.datanucleus.PropertyNames;
import org.datanucleus.identity.IdentityUtils;
import org.datanucleus.identity.OID;
import org.datanucleus.identity.OIDFactory;
import org.datanucleus.identity.SCOID;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.EmbeddedMetaData;
import org.datanucleus.metadata.IdentityStrategy;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.metadata.Relation;
import org.datanucleus.metadata.VersionMetaData;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.FieldValues;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.mongodb.fieldmanager.FetchFieldManager;
import org.datanucleus.store.schema.naming.ColumnType;
import org.datanucleus.util.NucleusLogger;

import com.mongodb.BasicDBObject;
import com.mongodb.DB;
import com.mongodb.DBCollection;
import com.mongodb.DBCursor;
import com.mongodb.DBObject;
import com.mongodb.MongoException;
import java.util.LinkedList;

/**
 * Utilities for MongoDB.
 */
public class MongoDBUtils
{
    public static List<Long> performMongoCount(DB db, BasicDBObject filterObject, Class candidateClass,
        boolean subclasses, ExecutionContext ec) 
    throws MongoException 
    {
        StoreManager storeMgr = ec.getStoreManager();
        long count = 0;
        for (AbstractClassMetaData cmd : MetaDataUtils.getMetaDataForCandidates(candidateClass, subclasses, ec)) 
        {
            String collectionName = storeMgr.getNamingFactory().getTableName(cmd);
            count += db.getCollection(collectionName).count(filterObject);
        }
        List<Long> results =  new LinkedList<Long>();
        results.add(count);
        if (ec.getStatistics() != null)
        {
            // Add to statistics
            ec.getStatistics().incrementNumReads();
        }
        return results;
    }

    /**
     * Accessor for the default value specified for the provided member.
     * If no defaultValue is provided on the column then returns null.
     * @param mmd Metadata for the member
     * @return The default value
     */
    public static String getDefaultValueForMember(AbstractMemberMetaData mmd)
    {
        ColumnMetaData[] colmds = mmd.getColumnMetaData();
        if (colmds == null || colmds.length < 1)
        {
            return null;
        }
        return colmds[0].getDefaultValue();
    }

    /**
     * Accessor for the MongoDB field for the field of this embedded field.
     * Uses the column name from the embedded definition (if present), otherwise falls back to
     * the column name of the field in its own class definition, otherwise uses its field name.
     * @param mmd Metadata for the owning member
     * @param fieldNumber Member number of the embedded object
     * @return The field name to use
     */
    public static String getFieldName(AbstractMemberMetaData mmd, int fieldNumber)
    {
        String columnName = null;

        // Try the first column if specified
        EmbeddedMetaData embmd = mmd.getEmbeddedMetaData();
        AbstractMemberMetaData embMmd = null;
        if (embmd != null)
        {
            AbstractMemberMetaData[] embmmds = embmd.getMemberMetaData();
            embMmd = embmmds[fieldNumber];
        }

        ColumnMetaData[] colmds = embMmd.getColumnMetaData();
        if (colmds != null && colmds.length > 0)
        {
            columnName = colmds[0].getName();
        }
        if (columnName == null)
        {
            // Fallback to the field/property name
            columnName = embMmd.getName();
        }
        if (columnName == null)
        {
            columnName = embMmd.getName();
        }
        return columnName;
    }

    /**
     * Convenience method that tries to find the object with the specified identity from all DBCollection objects
     * from the rootCmd and subclasses. Returns the class name of the object with this identity (or null if not found).
     * @param id The identity
     * @param rootCmd ClassMetaData for the root class in the inheritance tree
     * @param ec ExecutionContext
     * @param clr ClassLoader resolver
     * @return The class name of the object with this id
     */
    public static String getClassNameForIdentity(Object id, AbstractClassMetaData rootCmd, ExecutionContext ec,
            ClassLoaderResolver clr)
    {
        Map<String, Set<String>> classNamesByDbCollectionName = new HashMap<String, Set<String>>();
        StoreManager storeMgr = ec.getStoreManager();
        Set rootClassNames = new HashSet<String>();
        rootClassNames.add(rootCmd.getFullClassName());
        classNamesByDbCollectionName.put(storeMgr.getNamingFactory().getTableName(rootCmd), rootClassNames);
        String[] subclassNames = ec.getMetaDataManager().getSubclassesForClass(rootCmd.getFullClassName(), true);
        if (subclassNames != null)
        {
            for (int i=0;i<subclassNames.length;i++)
            {
                AbstractClassMetaData cmd = ec.getMetaDataManager().getMetaDataForClass(subclassNames[i], clr);
                String collName = storeMgr.getNamingFactory().getTableName(cmd);
                Set<String> classNames = classNamesByDbCollectionName.get(collName);
                if (classNames == null)
                {
                    classNames = new HashSet<String>();
                    classNamesByDbCollectionName.put(collName, classNames);
                }
                classNames.add(cmd.getFullClassName());
            }
        }

        ManagedConnection mconn = storeMgr.getConnection(ec);
        try
        {
            DB db = (DB)mconn.getConnection();

            for (String dbCollName : classNamesByDbCollectionName.keySet())
            {
                // Check each DBCollection for the id PK field(s)
                Set<String> classNames = classNamesByDbCollectionName.get(dbCollName);
                DBCollection dbColl = db.getCollection(dbCollName);
                BasicDBObject query = new BasicDBObject();
                if (id instanceof OID)
                {
                    // TODO Really ought to use cmd of the class using this dbCollection
                    Object key = ((OID)id).getKeyValue();
                    if (rootCmd.getIdentityMetaData().getValueStrategy() == IdentityStrategy.IDENTITY)
                    {
                        query.put("_id", new ObjectId((String)key));
                    }
                    else
                    {
                        query.put(storeMgr.getNamingFactory().getColumnName(rootCmd, ColumnType.DATASTOREID_COLUMN), key);
                    }
                }
                else if (ec.getApiAdapter().isSingleFieldIdentity(id))
                {
                    Object key = ec.getApiAdapter().getTargetKeyForSingleFieldIdentity(id);
                    int[] pkNums = rootCmd.getPKMemberPositions();
                    AbstractMemberMetaData pkMmd = rootCmd.getMetaDataForManagedMemberAtAbsolutePosition(pkNums[0]);
                    String pkPropName = storeMgr.getNamingFactory().getColumnName(pkMmd, ColumnType.COLUMN);
                    query.put(pkPropName, key);
                }
                else
                {
                    int[] pkNums = rootCmd.getPKMemberPositions();
                    for (int i=0;i<pkNums.length;i++)
                    {
                        AbstractMemberMetaData pkMmd = rootCmd.getMetaDataForManagedMemberAtAbsolutePosition(pkNums[i]);
                        String pkPropName = storeMgr.getNamingFactory().getColumnName(pkMmd, ColumnType.COLUMN);
                        Object pkVal = IdentityUtils.getValueForMemberInId(id, pkMmd);
                        query.put(pkPropName, pkVal);
                    }
                }

                if (NucleusLogger.DATASTORE_NATIVE.isDebugEnabled())
                {
                    NucleusLogger.DATASTORE_NATIVE.debug("Retrieving object for " + query + " from DBCollection with name " + dbCollName);
                }
                DBObject foundObj = dbColl.findOne(query);
                if (ec.getStatistics() != null)
                {
                    // Add to statistics
                    ec.getStatistics().incrementNumReads();
                }

                if (foundObj != null)
                {
                    if (classNames.size() == 1)
                    {
                        // Only one candidate so return it
                        return classNames.iterator().next();
                    }
                    else
                    {
                        String disPropName = storeMgr.getNamingFactory().getColumnName(rootCmd, ColumnType.DISCRIMINATOR_COLUMN);
                        String discValue = (String)foundObj.get(disPropName);
                        return MetaDataUtils.getClassNameFromDiscriminatorValue(discValue, rootCmd.getDiscriminatorMetaData(), ec);
                    }
                }
            }
        }
        finally
        {
            mconn.release();
        }
        return null;
    }

    /**
     * Method to return the DBObject that equates to the provided object.
     * @param dbCollection The collection in which it is stored
     * @param op The ObjectProvider
     * @param checkVersion Whether to also check for a particular version
     * @param originalValue Whether to use the original value of fields (when using nondurable id and doing update).
     * @return The object (or null if not found)
     */
    public static DBObject getObjectForObjectProvider(DBCollection dbCollection, ObjectProvider op, boolean checkVersion,
            boolean originalValue)
    {
        // Build query object to use as template for the find
        BasicDBObject query = new BasicDBObject();
        AbstractClassMetaData cmd = op.getClassMetaData();
        StoreManager storeMgr = op.getExecutionContext().getStoreManager();

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            // Application id - Add PK field(s) to the query object
            int[] pkPositions = cmd.getPKMemberPositions();
            for (int i=0;i<pkPositions.length;i++)
            {
                AbstractMemberMetaData pkMmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(pkPositions[i]);
                Object value = op.provideField(pkPositions[i]);
                if (value == null && pkMmd.getValueStrategy() == IdentityStrategy.IDENTITY)
                {
                    // PK field not yet set, so return null (needs to be attributed in the datastore)
                    return null;
                }
                if (pkMmd.getValueStrategy() == IdentityStrategy.IDENTITY)
                {
                    query.put("_id", new ObjectId((String)value));
                }
                else
                {
                    query.put(storeMgr.getNamingFactory().getColumnName(pkMmd, ColumnType.COLUMN), value);
                }
            }
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            // Datastore id - Add "id" field to the query object
            OID oid = (OID) op.getObjectId();
            if (oid == null && cmd.getIdentityMetaData().getValueStrategy() == IdentityStrategy.IDENTITY)
            {
                // Not yet set, so return null (needs to be attributed in the datastore)
                return null;
            }
            Object value = oid.getKeyValue();
            if (cmd.getIdentityMetaData().getValueStrategy() == IdentityStrategy.IDENTITY)
            {
                query.put("_id", new ObjectId((String)value));
            }
            else
            {
                query.put(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DATASTOREID_COLUMN), value);
            }
        }
        else
        {
            // Nondurable - Add all basic field(s) to the query object
            int[] fieldNumbers = cmd.getAllMemberPositions();
            for (int i=0;i<fieldNumbers.length;i++)
            {
                AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumbers[i]);
                int relationType = mmd.getRelationType(op.getExecutionContext().getClassLoaderResolver());
                if (relationType == Relation.NONE)
                {
                    Object fieldValue = null;
                    if (originalValue)
                    {
                        Object oldValue = op.getAssociatedValue("FIELD_VALUE.ORIGINAL." + fieldNumbers[i]);
                        if (oldValue != null)
                        {
                            fieldValue = oldValue;
                        }
                        else
                        {
                            fieldValue = op.provideField(fieldNumbers[i]);
                        }
                    }
                    else
                    {
                        fieldValue = op.provideField(fieldNumbers[i]);
                    }
                    query.put(storeMgr.getNamingFactory().getColumnName(mmd, ColumnType.COLUMN), fieldValue);
                }
            }
        }

        if (cmd.hasDiscriminatorStrategy())
        {
            // Add discriminator to the query object
            query.put(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DISCRIMINATOR_COLUMN), cmd.getDiscriminatorValue());
        }
        if (checkVersion && cmd.isVersioned())
        {
            // Add version to the query object
            Object currentVersion = op.getTransactionalVersion();
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Version field in class
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                String fieldName = storeMgr.getNamingFactory().getColumnName(verMmd, ColumnType.COLUMN);
                query.put(fieldName, currentVersion);
            }
            else
            {
                // Surrogate version field
                String fieldName = storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN);
                query.put(fieldName, currentVersion);
            }
        }

        if (NucleusLogger.DATASTORE_NATIVE.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_NATIVE.debug("Retrieving object for " + query);
        }
        DBObject dbObj = dbCollection.findOne(query);
        if (op.getExecutionContext().getStatistics() != null)
        {
            // Add to statistics
            op.getExecutionContext().getStatistics().incrementNumReads();
        }
        return dbObj;
    }

    public static List getObjectsOfCandidateType(ExecutionContext ec, DB db, Class candidateClass, 
            boolean subclasses, boolean ignoreCache, FetchPlan fp, BasicDBObject filterObject,
            Map<String, Object> options)
    {
        return getObjectsOfCandidateType(ec, db, candidateClass, subclasses, ignoreCache, fp, filterObject, options, null, null);
    }
    
    /**
     * Convenience method to return all objects of the candidate type (optionally allowing subclasses).
     * @param ec Execution context
     * @param db Mongo DB
     * @param candidateClass Candidate
     * @param subclasses Whether to include subclasses
     * @param ignoreCache Ignore the cache?
     * @param fp FetchPlan when retrieving the objects
     * @param filterObject Optional filter object
     * @param options Set of options for controlling this query
     * @return List of all candidate objects
     */
    public static List getObjectsOfCandidateType(ExecutionContext ec, DB db, Class candidateClass, 
            boolean subclasses, boolean ignoreCache, FetchPlan fp, BasicDBObject filterObject,
            Map<String, Object> options, Integer skip, Integer limit)
    {
        List<AbstractClassMetaData> cmds = 
            MetaDataUtils.getMetaDataForCandidates(candidateClass, subclasses, ec);

        // TODO Allow for subclasses sharing the same table so just doing a single query but with discriminator filtering
        Iterator<AbstractClassMetaData> cmdIter = cmds.iterator();
        List results = new ArrayList();
        StoreManager storeMgr = ec.getStoreManager();
        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        while (cmdIter.hasNext())
        {
            AbstractClassMetaData cmd = cmdIter.next();
            fp.manageFetchPlanForClass(cmd);
            int[] fpMembers = fp.getFetchPlanForClass(cmd).getMemberNumbers();
            BasicDBObject fieldsSelection = new BasicDBObject();
            if (fpMembers != null && fpMembers.length > 0)
            {
                fieldsSelection = new BasicDBObject();
                for (int i=0;i<fpMembers.length;i++)
                {
                    AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(fpMembers[i]);
                    int relationType = mmd.getRelationType(clr);
                    if (mmd.isEmbedded() && Relation.isRelationSingleValued(relationType))
                    {
                        boolean nested = true;
                        String nestedStr = mmd.getValueForExtension("nested");
                        if (nestedStr != null && nestedStr.equalsIgnoreCase("false"))
                        {
                            nested = false;
                        }

                        if (nested)
                        {
                            // Nested Embedded field, so include field
                            String fieldName = storeMgr.getNamingFactory().getColumnName(mmd, ColumnType.COLUMN);
                            fieldsSelection.append(fieldName, 1);
                        }
                        else
                        {
                            // Flat Embedded field, so add all fields of sub-objects
                            selectAllFieldsOfEmbeddedObject(mmd, fieldsSelection, ec, clr);
                        }
                    }
                    else
                    {
                        String fieldName = storeMgr.getNamingFactory().getColumnName(mmd, ColumnType.COLUMN);
                        fieldsSelection.append(fieldName, 1);
                    }
                }
            }
            if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                fieldsSelection.append(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DATASTOREID_COLUMN), 1);
            }
            if (cmd.isVersioned())
            {
                fieldsSelection.append(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN), 1);
            }

            BasicDBObject query = new BasicDBObject();
            if (filterObject != null)
            {
                Iterator<Map.Entry<String, Object>> entryIter = filterObject.entrySet().iterator();
                while (entryIter.hasNext())
                {
                    Map.Entry<String, Object> entry = entryIter.next();
                    query.put(entry.getKey(), entry.getValue());
                }
            }

            if (cmd.hasDiscriminatorStrategy())
            {
                // Discriminator present : Add restriction on the discriminator value for this class
                query.put(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DISCRIMINATOR_COLUMN), cmd.getDiscriminatorValue());
            }

            if (storeMgr.getStringProperty(PropertyNames.PROPERTY_TENANT_ID) != null)
            {
                // Multitenancy discriminator present : Add restriction for this tenant
                if ("true".equalsIgnoreCase(cmd.getValueForExtension("multitenancy-disable")))
                {
                    // Don't bother with multitenancy for this class
                }
                else
                {
                    String fieldName = storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.MULTITENANCY_COLUMN);
                    String value = storeMgr.getStringProperty(PropertyNames.PROPERTY_TENANT_ID);
                    query.put(fieldName, value);
                }
            }

            String collectionName = storeMgr.getNamingFactory().getTableName(cmd);
            if (NucleusLogger.DATASTORE_RETRIEVE.isDebugEnabled())
            {
                NucleusLogger.DATASTORE_RETRIEVE.debug("Fetching instances of collection " + collectionName +
                    " fields=" + fieldsSelection + " with filter=" + query);
            }

            DBCollection dbColl = db.getCollection(collectionName);
            Object val = (options != null ? options.get("slave-ok") : Boolean.FALSE);
            if (val == Boolean.TRUE)
            {
                dbColl.slaveOk();
            }

            DBCursor curs = dbColl.find(query, fieldsSelection);
            if (ec.getStatistics() != null)
            {
                // Add to statistics
                ec.getStatistics().incrementNumReads();
            }

            // Add results from this DBCursor and close it
            // TODO Pass this into a MongoDBCandidateList rather than manually loading all here
            if (skip != null && skip > 0)
            {
                curs = curs.skip(skip);
            }
            if (limit != null && limit > 0)
            {
                curs = curs.limit(limit);
            }
            try
            {
                while (curs.hasNext())
                {
                    final DBObject dbObject = curs.next();
                    if (cmd.getIdentityType() == IdentityType.APPLICATION)
                    {
                        results.add(getObjectUsingApplicationIdForDBObject(dbObject, cmd, ec, ignoreCache, fpMembers));
                    }
                    else if (cmd.getIdentityType() == IdentityType.DATASTORE)
                    {
                        results.add(getObjectUsingDatastoreIdForDBObject(dbObject, cmd, ec, ignoreCache, fpMembers));
                    }
                    else
                    {
                        results.add(getObjectUsingNondurableIdForDBObject(dbObject, cmd, ec, ignoreCache, fpMembers));
                    }
                }
            }
            finally
            {
                curs.close();
            }
        }

        return results;
    }

    protected static void selectAllFieldsOfEmbeddedObject(AbstractMemberMetaData mmd, BasicDBObject fieldsSelection,
            ExecutionContext ec, ClassLoaderResolver clr)
    {
        EmbeddedMetaData embmd = mmd.getEmbeddedMetaData();
        AbstractMemberMetaData[] embmmds = embmd.getMemberMetaData();
        for (int i=0;i<embmmds.length;i++)
        {
            int relationType = embmmds[i].getRelationType(clr);
            if (embmmds[i].isEmbedded() && Relation.isRelationSingleValued(relationType))
            {
                selectAllFieldsOfEmbeddedObject(embmmds[i], fieldsSelection, ec, clr);
            }
            else
            {
                String embFieldName = MongoDBUtils.getFieldName(mmd, i);
                fieldsSelection.append(embFieldName, 1);
            }
        }
    }

    public static Object getObjectUsingApplicationIdForDBObject(final DBObject dbObject, final AbstractClassMetaData cmd,
            final ExecutionContext ec, boolean ignoreCache, final int[] fpMembers)
    {
        Object id = IdentityUtils.getApplicationIdentityForResultSetRow(ec, cmd, null, 
            false, new FetchFieldManager(ec, dbObject, cmd));

        StoreManager storeMgr = ec.getStoreManager();
        Class type = ec.getClassLoaderResolver().classForName(cmd.getFullClassName());
        Object pc = ec.findObject(id, 
            new FieldValues()
            {
                public void fetchFields(ObjectProvider sm)
                {
                    sm.replaceFields(fpMembers, new FetchFieldManager(ec, dbObject, cmd));
                }
                public void fetchNonLoadedFields(ObjectProvider sm)
                {
                    sm.replaceNonLoadedFields(fpMembers, new FetchFieldManager(ec, dbObject, cmd));
                }
                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, type, ignoreCache, false);

        if (cmd.isVersioned())
        {
            // Set the version on the retrieved object
            ObjectProvider sm = ec.findObjectProvider(pc);
            Object version = null;
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Get the version from the field value
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                version = sm.provideField(verMmd.getAbsoluteFieldNumber());
            }
            else
            {
                // Get the surrogate version from the datastore
                version = dbObject.get(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN));
            }
            sm.setVersion(version);
        }
        return pc;
    }

    public static Object getObjectUsingDatastoreIdForDBObject(final DBObject dbObject, final AbstractClassMetaData cmd,
            final ExecutionContext ec, boolean ignoreCache, final int[] fpMembers)
    {
        Object idKey = null;
        StoreManager storeMgr = ec.getStoreManager();
        if (cmd.getIdentityMetaData().getValueStrategy() == IdentityStrategy.IDENTITY)
        {
            idKey = dbObject.get("_id");
            if (idKey instanceof ObjectId)
            {
                idKey = ((ObjectId)idKey).toString();
            }
        }
        else
        {
            idKey = dbObject.get(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DATASTOREID_COLUMN));
        }
        OID oid = OIDFactory.getInstance(ec.getNucleusContext(), cmd.getFullClassName(), idKey);
        Class type = ec.getClassLoaderResolver().classForName(cmd.getFullClassName());
        Object pc = ec.findObject(oid, 
            new FieldValues()
            {
                // StateManager calls the fetchFields method
                public void fetchFields(ObjectProvider sm)
                {
                    sm.replaceFields(fpMembers, new FetchFieldManager(ec, dbObject, cmd));
                }
                public void fetchNonLoadedFields(ObjectProvider sm)
                {
                    sm.replaceNonLoadedFields(fpMembers, new FetchFieldManager(ec, dbObject, cmd));
                }
                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, type, ignoreCache, false);

        if (cmd.isVersioned())
        {
            // Set the version on the retrieved object
            ObjectProvider sm = ec.findObjectProvider(pc);
            Object version = null;
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Get the version from the field value
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                version = sm.provideField(verMmd.getAbsoluteFieldNumber());
            }
            else
            {
                // Get the surrogate version from the datastore
                version = dbObject.get(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN));
            }
            sm.setVersion(version);
        }
        return pc;
    }
    public static Object getObjectUsingNondurableIdForDBObject(final DBObject dbObject, final AbstractClassMetaData cmd,
            final ExecutionContext ec, boolean ignoreCache, final int[] fpMembers)
    {
        SCOID oid = new SCOID(cmd.getFullClassName());
        Class type = ec.getClassLoaderResolver().classForName(cmd.getFullClassName());
        Object pc = ec.findObject(oid, 
            new FieldValues()
            {
                // StateManager calls the fetchFields method
                public void fetchFields(ObjectProvider sm)
                {
                    sm.replaceFields(fpMembers, new FetchFieldManager(ec, dbObject, cmd));
                }
                public void fetchNonLoadedFields(ObjectProvider sm)
                {
                    sm.replaceNonLoadedFields(fpMembers, new FetchFieldManager(ec, dbObject, cmd));
                }
                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, type, ignoreCache, false);

        if (cmd.isVersioned())
        {
            // Set the version on the retrieved object
            ObjectProvider sm = ec.findObjectProvider(pc);
            StoreManager storeMgr = ec.getStoreManager();
            Object version = null;
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Get the version from the field value
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                version = sm.provideField(verMmd.getAbsoluteFieldNumber());
            }
            else
            {
                // Get the surrogate version from the datastore
                version = dbObject.get(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN));
            }
            sm.setVersion(version);
        }
        return pc;
    }
}